import os
import json
import gzip
import time
import hmac
import random
import asyncio
import hashlib

from urllib.parse import urljoin

import aiohttp

from aiohttp.client_exceptions import ContentTypeError


class ApiError(Exception):
    pass


class ApiProvider:
    """Provides HTTP-api requests and websocket requests."""
    def __init__(
            self, *, api_key='', api_secret='', from_env=False,
            auth_required=True, timeout=25, retries=6,
            root_url='https://api.crypto.com/v2/',
            ws_root_url='wss://stream.crypto.com/v2/', logger=None):
        self.api_key = api_key
        self.api_secret = api_secret
        self.root_url = root_url
        self.ws_root_url = ws_root_url
        self.timeout = timeout
        self.retries = retries

        # NOTE: do not change this, due to crypto.com rate-limits
        self.semaphore = asyncio.Semaphore(40)

        if not auth_required:
            return

        if from_env:
            self.read_keys_from_env()

        if not self.api_key or not self.api_secret:
            raise ValueError('Provide api_key and api_secret')

    def read_keys_from_env(self):
        self.api_key = os.environ.get('CRYPTOCOM_API_KEY', '')
        if not self.api_key:
            raise ValueError('Provide CRYPTOCOM_API_KEY env value')
        self.api_secret = os.environ.get('CRYPTOCOM_API_SECRET', '')
        if not self.api_secret:
            raise ValueError('Provide CRYPTOCOM_API_SECRET env value')

    def _sign(self, path, data):
        data = data or {}
        data['method'] = path

        sign_time = int(time.time() * 1000)
        data.update({'nonce': sign_time, 'api_key': self.api_key})

        data['id'] = random.randint(1000, 10000)
        data_params = data.get('params', {})
        params = ''.join(
            f'{key}{data_params[key]}'
            for key in sorted(data_params)
        )

        payload = f"{path}{data['id']}" \
            f"{self.api_key}{params}{data['nonce']}"

        data['sig'] = hmac.new(
            self.api_secret.encode('utf-8'),
            payload.encode('utf-8'),
            hashlib.sha256
        ).hexdigest()
        return data

    async def request(self, method, path, params=None, data=None, sign=False):
        original_data = data
        timeout = aiohttp.ClientTimeout(total=self.timeout)
        for count in range(self.retries + 1):
            if sign:
                data = self._sign(path, original_data)
            try:
                async with aiohttp.ClientSession(timeout=timeout) as session:
                    async with self.semaphore:
                        resp = await session.request(
                            method, urljoin(self.root_url, path),
                            params=params, json=data,
                            headers={'content-type': 'application/json'}
                        )
                        resp_json = await resp.json()
                        if resp.status != 200:
                            raise ApiError(
                                f"Error: {resp_json}. "
                                f"Status: {resp.status}. Json params: {data}")
            except aiohttp.ClientConnectorError:
                raise ApiError(f"Cannot connect to host {self.root_url}")
            except asyncio.TimeoutError:
                if count == self.retries:
                    raise ApiError(
                        f"Timeout error, retries: {self.retries}. "
                        f"Code: {resp.status}. Data: {data}"
                    )

                await asyncio.sleep(0.5)
                continue
            except ContentTypeError:
                if resp.status == 429:
                    await asyncio.sleep(0.5)
                    continue
                text = await resp.text()
                raise ApiError(
                    f"Can't decode json, content: {text}. "
                    f"Code: {resp.status}")

            if resp_json['code'] == 0:
                result = resp_json.get('result', {})
                return result.get('data', {}) or result

            if count == self.retries:
                raise ApiError(
                    f"System error, retries: {self.retries}. "
                    f"Code: {resp.status}. Json: {resp_json}. "
                    f"Data: {data}"
                )

            await asyncio.sleep(0.5)
            continue

    async def get(self, path, params=None, sign=False):
        return await self.request('get', path, params=params, sign=sign)

    async def post(self, path, data=None, sign=True):
        return await self.request('post', path, data=data, sign=sign)

    async def listen(self, url, *channels, sign=False):
        while True:
            try:
                async for data in self.listen_once(url, *channels, sign=sign):
                    yield data
            except Exception:
                await asyncio.sleep(1)

    async def listen_once(self, url, *channels, sign=False):
        timeout = aiohttp.ClientTimeout(10)
        url = urljoin(self.ws_root_url, url)
        sub_data = {
            'id': random.randint(1000, 10000),
            "method": "subscribe",
            'params': {"channels": list(channels)},
            "nonce": int(time.time())
        }
        auth_data = {}
        if sign:
            auth_data = self._sign('public/auth', auth_data)

        sub_data_sent = False
        auth_sent = False

        async with aiohttp.ClientSession(timeout=timeout) as session:
            async with session.ws_connect(url) as ws:
                # sleep because too many requests from docs
                await asyncio.sleep(1)

                # [0] if not sign start connection with subscription
                if not sign:
                    await ws.send_str(json.dumps(sub_data))
                    sub_data_sent = True

                async for msg in ws:
                    data = json.loads(msg.data)
                    result = data.get('result')

                    # [1] send heartbeat to keep connection alive
                    if data['method'] == 'public/heartbeat':
                        await ws.send_str(json.dumps({
                            'id': data['id'],
                            'method': 'public/respond-heartbeat'
                        }))
                    elif sub_data_sent:
                        # [4] consume data
                        if result:
                            yield result
                        continue

                    # [2] sign auth request to listen private methods
                    if sign and not auth_sent:
                        await ws.send_str(json.dumps(auth_data))
                        auth_sent = True

                    # [3] subscribe to channels
                    if (
                            data['method'] == 'public/auth' and
                            data['code'] == 0 and not sub_data_sent
                    ):
                        await ws.send_str(json.dumps(sub_data))
                        sub_data_sent = True



from .structs import Pair

AAVE_CRO = Pair("AAVE_CRO", price_precision=2, quantity_precision=3)
AAVE_USDT = Pair("AAVE_USDT", price_precision=3, quantity_precision=4)
ADA_BTC = Pair("ADA_BTC", price_precision=9, quantity_precision=0)
ADA_CRO = Pair("ADA_CRO", price_precision=3, quantity_precision=2)
ADA_USDT = Pair("ADA_USDT", price_precision=5, quantity_precision=1)
ALGO_BTC = Pair("ALGO_BTC", price_precision=8, quantity_precision=2)
ALGO_CRO = Pair("ALGO_CRO", price_precision=3, quantity_precision=2)
ALGO_USDT = Pair("ALGO_USDT", price_precision=4, quantity_precision=2)
ATOM_BTC = Pair("ATOM_BTC", price_precision=7, quantity_precision=2)
ATOM_CRO = Pair("ATOM_CRO", price_precision=2, quantity_precision=2)
ATOM_USDT = Pair("ATOM_USDT", price_precision=3, quantity_precision=2)
BAND_CRO = Pair("BAND_CRO", price_precision=3, quantity_precision=2)
BAND_USDT = Pair("BAND_USDT", price_precision=4, quantity_precision=2)
BAT_CRO = Pair("BAT_CRO", price_precision=3, quantity_precision=2)
BAT_USDT = Pair("BAT_USDT", price_precision=4, quantity_precision=2)
BCH_BTC = Pair("BCH_BTC", price_precision=6, quantity_precision=3)
BCH_CRO = Pair("BCH_CRO", price_precision=2, quantity_precision=3)
BCH_USDT = Pair("BCH_USDT", price_precision=2, quantity_precision=5)
BNT_CRO = Pair("BNT_CRO", price_precision=3, quantity_precision=2)
BNT_USDT = Pair("BNT_USDT", price_precision=4, quantity_precision=3)
BTC_USDC = Pair("BTC_USDC", price_precision=2, quantity_precision=6)
BTC_USDT = Pair("BTC_USDT", price_precision=2, quantity_precision=6)
CELR_CRO = Pair("CELR_CRO", price_precision=5, quantity_precision=1)
CELR_USDT = Pair("CELR_USDT", price_precision=6, quantity_precision=1)
COMP_CRO = Pair("COMP_CRO", price_precision=2, quantity_precision=3)
COMP_USDT = Pair("COMP_USDT", price_precision=2, quantity_precision=5)
CRO_BTC = Pair("CRO_BTC", price_precision=9, quantity_precision=2)
CRO_USDC = Pair("CRO_USDC", price_precision=5, quantity_precision=3)
CRO_USDT = Pair("CRO_USDT", price_precision=5, quantity_precision=3)
CRV_CRO = Pair("CRV_CRO", price_precision=3, quantity_precision=2)
CRV_USDT = Pair("CRV_USDT", price_precision=4, quantity_precision=4)
DAI_CRO = Pair("DAI_CRO", price_precision=3, quantity_precision=3)
DAI_USDT = Pair("DAI_USDT", price_precision=4, quantity_precision=3)
DOT_CRO = Pair("DOT_CRO", price_precision=3, quantity_precision=3)
DOT_USDT = Pair("DOT_USDT", price_precision=4, quantity_precision=3)
EGLD_CRO = Pair("EGLD_CRO", price_precision=3, quantity_precision=2)
EGLD_USDT = Pair("EGLD_USDT", price_precision=4, quantity_precision=3)
ENJ_CRO = Pair("ENJ_CRO", price_precision=3, quantity_precision=2)
ENJ_USDT = Pair("ENJ_USDT", price_precision=5, quantity_precision=1)
EOS_BTC = Pair("EOS_BTC", price_precision=8, quantity_precision=2)
EOS_CRO = Pair("EOS_CRO", price_precision=3, quantity_precision=2)
EOS_USDT = Pair("EOS_USDT", price_precision=4, quantity_precision=2)
ETH_BTC = Pair("ETH_BTC", price_precision=6, quantity_precision=5)
ETH_CRO = Pair("ETH_CRO", price_precision=2, quantity_precision=5)
ETH_USDT = Pair("ETH_USDT", price_precision=2, quantity_precision=5)
GRT_CRO = Pair("GRT_CRO", price_precision=3, quantity_precision=2)
GRT_USDT = Pair("GRT_USDT", price_precision=5, quantity_precision=2)
ICX_BTC = Pair("ICX_BTC", price_precision=8, quantity_precision=1)
ICX_CRO = Pair("ICX_CRO", price_precision=3, quantity_precision=0)
ICX_USDT = Pair("ICX_USDT", price_precision=4, quantity_precision=2)
KNC_CRO = Pair("KNC_CRO", price_precision=3, quantity_precision=2)
KNC_USDT = Pair("KNC_USDT", price_precision=4, quantity_precision=3)
KSM_CRO = Pair("KSM_CRO", price_precision=2, quantity_precision=3)
KSM_USDT = Pair("KSM_USDT", price_precision=3, quantity_precision=3)
LINK_BTC = Pair("LINK_BTC", price_precision=7, quantity_precision=2)
LINK_CRO = Pair("LINK_CRO", price_precision=2, quantity_precision=2)
LINK_USDT = Pair("LINK_USDT", price_precision=3, quantity_precision=2)
LRC_CRO = Pair("LRC_CRO", price_precision=4, quantity_precision=2)
LRC_USDT = Pair("LRC_USDT", price_precision=5, quantity_precision=2)
LTC_BTC = Pair("LTC_BTC", price_precision=6, quantity_precision=4)
LTC_CRO = Pair("LTC_CRO", price_precision=2, quantity_precision=4)
LTC_USDT = Pair("LTC_USDT", price_precision=2, quantity_precision=5)
MANA_BTC = Pair("MANA_BTC", price_precision=9, quantity_precision=1)
MANA_CRO = Pair("MANA_CRO", price_precision=3, quantity_precision=2)
MANA_USDT = Pair("MANA_USDT", price_precision=5, quantity_precision=2)
MKR_CRO = Pair("MKR_CRO", price_precision=2, quantity_precision=3)
MKR_USDT = Pair("MKR_USDT", price_precision=2, quantity_precision=6)
NEAR_CRO = Pair("NEAR_CRO", price_precision=3, quantity_precision=2)
NEAR_USDT = Pair("NEAR_USDT", price_precision=4, quantity_precision=2)
NEO_BTC = Pair("NEO_BTC", price_precision=7, quantity_precision=3)
NEO_CRO = Pair("NEO_CRO", price_precision=3, quantity_precision=2)
NEO_USDT = Pair("NEO_USDT", price_precision=3, quantity_precision=4)
OMG_BTC = Pair("OMG_BTC", price_precision=8, quantity_precision=4)
OMG_CRO = Pair("OMG_CRO", price_precision=3, quantity_precision=2)
OMG_USDT = Pair("OMG_USDT", price_precision=4, quantity_precision=4)
ONT_BTC = Pair("ONT_BTC", price_precision=8, quantity_precision=2)
ONT_CRO = Pair("ONT_CRO", price_precision=3, quantity_precision=2)
ONT_USDT = Pair("ONT_USDT", price_precision=4, quantity_precision=2)
PAXG_CRO = Pair("PAXG_CRO", price_precision=2, quantity_precision=4)
PAXG_USDT = Pair("PAXG_USDT", price_precision=2, quantity_precision=5)
QTUM_CRO = Pair("QTUM_CRO", price_precision=3, quantity_precision=2)
QTUM_USDT = Pair("QTUM_USDT", price_precision=4, quantity_precision=4)
REN_CRO = Pair("REN_CRO", price_precision=4, quantity_precision=2)
REN_USDT = Pair("REN_USDT", price_precision=5, quantity_precision=2)
SKL_USDT = Pair("SKL_USDT", price_precision=5, quantity_precision=2)
UNI_CRO = Pair("UNI_CRO", price_precision=3, quantity_precision=2)
UNI_USDT = Pair("UNI_USDT", price_precision=4, quantity_precision=4)
USDC_USDT = Pair("USDC_USDT", price_precision=4, quantity_precision=2)
VET_BTC = Pair("VET_BTC", price_precision=10, quantity_precision=0)
VET_CRO = Pair("VET_CRO", price_precision=4, quantity_precision=0)
VET_USDT = Pair("VET_USDT", price_precision=6, quantity_precision=0)
WBTC_BTC = Pair("WBTC_BTC", price_precision=4, quantity_precision=6)
WBTC_USDT = Pair("WBTC_USDT", price_precision=2, quantity_precision=6)
XLM_BTC = Pair("XLM_BTC", price_precision=9, quantity_precision=0)
XLM_CRO = Pair("XLM_CRO", price_precision=3, quantity_precision=2)
XLM_USDT = Pair("XLM_USDT", price_precision=5, quantity_precision=1)
XRP_BTC = Pair("XRP_BTC", price_precision=9, quantity_precision=0)
XRP_CRO = Pair("XRP_CRO", price_precision=3, quantity_precision=0)
XRP_USDT = Pair("XRP_USDT", price_precision=5, quantity_precision=1)
XTZ_BTC = Pair("XTZ_BTC", price_precision=8, quantity_precision=2)
XTZ_CRO = Pair("XTZ_CRO", price_precision=3, quantity_precision=2)
XTZ_USDT = Pair("XTZ_USDT", price_precision=4, quantity_precision=2)
YFI_BTC = Pair("YFI_BTC", price_precision=4, quantity_precision=6)
YFI_CRO = Pair("YFI_CRO", price_precision=2, quantity_precision=6)
YFI_USDT = Pair("YFI_USDT", price_precision=2, quantity_precision=6)
ZIL_CRO = Pair("ZIL_CRO", price_precision=4, quantity_precision=2)
ZIL_USDT = Pair("ZIL_USDT", price_precision=5, quantity_precision=1)


def all():
    return [
        value for name, value in globals().items()
        if isinstance(value, Pair)
    ]



import asyncio

from typing import List, Dict

from .api import ApiProvider, ApiError
from .market import Exchange
from .structs import (
    Pair, OrderSide, OrderStatus, OrderType, Order, Coin, Balance,
    OrderExecType, OrderForceType, PrivateTrade
)


class Account:
    """Provides access to account actions and data. Balance, trades, orders."""
    def __init__(
            self, *, api_key: str = '', api_secret: str = '',
            from_env: bool = False, exchange: Exchange = None,
            api: ApiProvider = None):
        if not api and not (api_key and api_secret) and not from_env:
            raise ValueError(
                'Pass ApiProvider or api_key with api_secret or from_env')
        self.api = api or ApiProvider(
            api_key=api_key, api_secret=api_secret, from_env=from_env)
        self.exchange = exchange or Exchange(api)
        self.pairs = self.exchange.pairs

    async def sync_pairs(self):
        await self.exchange.sync_pairs()
        self.pairs = self.exchange.pairs

    async def get_balance(self) -> Dict[Coin, Balance]:
        """Return balance."""
        data = await self.api.post('private/get-account-summary')
        return {
            Coin(bal['currency']): Balance.from_api(bal)
            for bal in data['accounts']
        }

    async def get_orders_history(
            self, pair: Pair = None, page: int = 0,
            page_size: int = 200) -> List[Order]:
        """Return all history orders."""
        params = {'page_size': page_size, 'page': page}
        if pair:
            params['instrument_name'] = pair.name
        data = await self.api.post(
            'private/get-order-history', {'params': params})
        return [
            Order.create_from_api(self.pairs[order['instrument_name']], order)
            for order in data.get('order_list') or []
        ]

    async def get_open_orders(
            self, pair: Pair = None, page: int = 0,
            page_size: int = 200) -> List[Order]:
        """Return open orders."""
        params = {'page_size': page_size, 'page': page}
        if pair:
            params['instrument_name'] = pair.name
        data = await self.api.post(
            'private/get-open-orders', {'params': params})
        return [
            Order.create_from_api(self.pairs[order['instrument_name']], order)
            for order in data.get('order_list') or []
        ]

    async def get_trades(
            self, pair: Pair = None, page: int = 0,
            page_size: int = 200) -> List[PrivateTrade]:
        """Return trades."""
        params = {'page_size': page_size, 'page': page}
        if pair:
            params['instrument_name'] = pair.name
        data = await self.api.post('private/get-trades', {'params': params})
        return [
            PrivateTrade.create_from_api(
                self.pairs[trade['instrument_name']], trade
            )
            for trade in data.get('trade_list') or []
        ]

    async def create_order(
            self, pair: Pair, side: OrderSide, type_: OrderType,
            quantity: float, price: float = 0,
            force_type: OrderForceType = OrderForceType.GOOD_TILL_CANCEL,
            exec_type: OrderExecType = OrderExecType.MARKET,
            client_id: int = None) -> int:
        """Create raw order with buy or sell side."""
        data = {
            'instrument_name': pair.name, 'side': side.value,
            'type': type_.value
        }
        data['time_in_force'] = force_type.value
        data['exec_inst'] = exec_type.value

        quantity = "{:.{}f}".format(quantity, pair.quantity_precision)
        if type_ == OrderType.MARKET and side == OrderSide.BUY:
            data['notional'] = quantity
        else:
            data['quantity'] = quantity

        if client_id:
            data['client_oid'] = str(client_id)

        if price:
            if type_ == OrderType.MARKET:
                raise ValueError(
                    "Error, MARKET execution do not support price value")
            data['price'] = "{:.{}f}".format(price, pair.price_precision)

        resp = await self.api.post('private/create-order', {'params': data})
        return int(resp['order_id'])

    async def buy_limit(
            self, pair: Pair, quantity: float, price: float,
            force_type: OrderForceType = OrderForceType.GOOD_TILL_CANCEL,
            exec_type: OrderExecType = OrderExecType.MARKET,
            client_id: int = None) -> int:
        """Buy limit order."""
        return await self.create_order(
            pair, OrderSide.BUY, OrderType.LIMIT, quantity, price,
            force_type, exec_type
        )

    async def sell_limit(
            self, pair: Pair, quantity: float, price: float,
            force_type: OrderForceType = OrderForceType.GOOD_TILL_CANCEL,
            exec_type: OrderExecType = OrderExecType.MARKET,
            client_id: int = None) -> int:
        """Sell limit order."""
        return await self.create_order(
            pair, OrderSide.SELL, OrderType.LIMIT, quantity, price,
            force_type, exec_type
        )

    async def wait_for_status(
            self, order_id: int, statuses, delay: int = 0.1) -> None:
        """Wait for order status."""
        order = await self.get_order(order_id)

        for _ in range(self.api.retries):
            if order.status in statuses:
                break

            await asyncio.sleep(delay)
            order = await self.get_order(order_id)

        if order.status not in statuses:
            raise ApiError(
                f"Status not changed for: {order}, must be in: {statuses}")

    async def buy_market(
            self, pair: Pair, spend: float, wait_for_fill=False) -> int:
        """Buy market order."""
        order_id = await self.create_order(
            pair, OrderSide.BUY, OrderType.MARKET, spend
        )
        if wait_for_fill:
            await self.wait_for_status(order_id, (
                OrderStatus.FILLED, OrderStatus.CANCELED, OrderStatus.EXPIRED,
                OrderStatus.REJECTED
            ))

        return order_id

    async def sell_market(
            self, pair: Pair, quantity: float, wait_for_fill=False) -> int:
        """Sell market order."""
        order_id = await self.create_order(
            pair, OrderSide.SELL, OrderType.MARKET, quantity
        )

        if wait_for_fill:
            await self.wait_for_status(order_id, (
                OrderStatus.FILLED, OrderStatus.CANCELED, OrderStatus.EXPIRED,
                OrderStatus.REJECTED
            ))

        return order_id

    async def get_order(self, order_id: int) -> Order:
        """Get order info."""
        order_info = {}
        retries = 0

        while True:
            data = await self.api.post('private/get-order-detail', {
                'params': {'order_id': str(order_id)}
            })
            order_info = data.get('order_info', {})
            if not order_info and retries < 10:
                await asyncio.sleep(0.5)
                retries += 1
            else:
                break

        return Order.create_from_api(
            self.pairs[order_info['instrument_name']],
            order_info, data['trade_list']
        )

    async def cancel_order(
            self, order_id: int, pair: Pair, check_status=False) -> None:
        """Cancel order."""
        await self.api.post('private/cancel-order', {
            'params': {'order_id': order_id, 'instrument_name': pair.name}
        })

        if not check_status:
            return

        await self.wait_for_status(order_id, (
            OrderStatus.CANCELED, OrderStatus.EXPIRED, OrderStatus.REJECTED
        ))

    async def cancel_open_orders(self, pair: Pair) -> None:
        """Cancel all open orders."""
        await self.api.post('private/cancel-all-orders', {
            'params': {'instrument_name': pair.name}
        })

    async def listen_balance(self) -> Balance:
        async for data in self.api.listen(
                'user', 'user.balance', sign=True):
            for bal in data.get('data', []):
                yield Balance(
                    total=bal['balance'],
                    available=bal['available'],
                    in_orders=bal['order'],
                    in_stake=bal['stake'],
                    coin=Coin(bal['currency'])
                )

    async def listen_orders(self, pair: Pair) -> Order:
        async for data in self.api.listen(
                'user', f'user.order.{pair.name}', sign=True):
            for order in data.get('data', []):
                yield Order.create_from_api(
                    self.pairs[data['instrument_name']], order
                )



from pathlib import Path
from setuptools import setup, find_packages


def get_version():
    module_file = Path(
        Path(__file__).parent, 'src',
        *find_packages('src')[-1].split('.'), '__init__.py'
    )
    return module_file.read_text().split("__version__ = '")[1].split("'")[0]


setup(
    name='cryptocom-exchange',
    version=get_version(),
    description="""Provide description""",
    url='https://github.com/goincrypto/cryptocom-exchange',
    author='Yaroslav Rudenok [Morty Space]',
    author_email='goincrypto@gmail.com',
    license='MIT',
    classifiers=[
        # TODO: add classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        'aiohttp',
        'cached-property'
    ],
    extras_require={
        'dev': [
            'pylint',
            'pycodestyle',
            'pytest',
            'pytest-asyncio',
            'pytest-cov',
            'pytest-env',
            'pytest-doctestplus',
            'pytest-pythonpath',
            'autopep8',
            'sphinx',
            'sphinx_rtd_theme',
            'setuptools',
            'twine',
            'doc8'
        ]
    },
    zip_safe=True,
    long_description=(Path(__file__).parent / 'README.md').read_text(),
    long_description_content_type='text/markdown'
)



from .structs import Coin

AAVE = Coin("AAVE")
ADA = Coin("ADA")
ALGO = Coin("ALGO")
ATOM = Coin("ATOM")
BAND = Coin("BAND")
BAT = Coin("BAT")
BCH = Coin("BCH")
BNT = Coin("BNT")
BTC = Coin("BTC")
CELR = Coin("CELR")
COMP = Coin("COMP")
CRO = Coin("CRO")
CRV = Coin("CRV")
DAI = Coin("DAI")
DOT = Coin("DOT")
EGLD = Coin("EGLD")
ENJ = Coin("ENJ")
EOS = Coin("EOS")
ETC = Coin("ETC")
ETH = Coin("ETH")
GRT = Coin("GRT")
ICX = Coin("ICX")
KNC = Coin("KNC")
KSM = Coin("KSM")
LINK = Coin("LINK")
LRC = Coin("LRC")
LTC = Coin("LTC")
MANA = Coin("MANA")
MCO = Coin("MCO")
MKR = Coin("MKR")
NEAR = Coin("NEAR")
NEO = Coin("NEO")
OMG = Coin("OMG")
ONT = Coin("ONT")
PAXG = Coin("PAXG")
QTUM = Coin("QTUM")
REN = Coin("REN")
SKL = Coin("SKL")
STX = Coin("STX")
UNI = Coin("UNI")
USDC = Coin("USDC")
USDT = Coin("USDT")
VET = Coin("VET")
WBTC = Coin("WBTC")
XLM = Coin("XLM")
XRP = Coin("XRP")
XTZ = Coin("XTZ")
YFI = Coin("YFI")
ZIL = Coin("ZIL")


def all():
    return [
        value for name, value in globals().items()
        if isinstance(value, Coin)
    ]



import time

from enum import Enum
from typing import List, Dict
from dataclasses import dataclass

from cached_property import cached_property

from .helpers import round_down, round_up


@dataclass
class Coin:
    name: str

    def __hash__(self):
        return self.name.__hash__()


@dataclass
class Pair:
    name: str
    price_precision: int
    quantity_precision: int

    @cached_property
    def base_coin(self) -> Coin:
        return Coin(self.name.split('_')[0])

    @cached_property
    def quote_coin(self) -> Coin:
        return Coin(self.name.split('_')[1])

    def round_price(self, price):
        return round_down(price, self.price_precision)

    def round_quantity(self, quantity):
        return round_down(quantity, self.quantity_precision)

    def __hash__(self):
        return self.name.__hash__()


@dataclass
class MarketTicker:
    pair: Pair
    buy_price: float
    sell_price: float
    trade_price: float
    time: int
    volume: float
    high: float
    low: float
    change: float

    @classmethod
    def from_api(cls, pair, data):
        return cls(
            pair=pair,
            buy_price=pair.round_price(data['b']),
            sell_price=pair.round_price(data['k']),
            trade_price=pair.round_price(data['a']),
            time=int(data['t'] / 1000),
            volume=pair.round_quantity(data['v']),
            high=pair.round_price(data['h']),
            low=pair.round_price(data['l']),
            change=round_down(data['c'], 3)
        )


class OrderSide(str, Enum):
    BUY = 'BUY'
    SELL = 'SELL'


@dataclass
class MarketTrade:
    id: int
    time: int
    price: float
    quantity: float
    side: OrderSide
    pair: Pair

    @classmethod
    def from_api(cls, pair: Pair, data: Dict):
        return cls(
            id=data['d'],
            time=int(data['t'] / 1000),
            price=pair.round_price(data['p']),
            quantity=pair.round_quantity(data['q']),
            side=OrderSide(data['s'].upper()),
            pair=pair
        )


class Period(str, Enum):
    MINS = '1m'
    MINS_5 = '5m'
    MINS_15 = '15m'
    MINS_30 = '30m'
    HOURS = '1h'
    HOURS_4 = '4h'
    HOURS_6 = '6h'
    HOURS_12 = '12h'
    DAY = '1D'
    WEEK = '7D'
    WEEK_2 = '14D'
    MONTH_1 = '1M'


@dataclass
class Candle:
    time: int
    open: float
    high: float
    low: float
    close: float
    volume: float
    pair: Pair

    @classmethod
    def from_api(cls, pair: Pair, data: Dict):
        return cls(
            time=int(data['t'] / 1000),
            open=pair.round_price(data['o']),
            high=pair.round_price(data['h']),
            low=pair.round_price(data['l']),
            close=pair.round_price(data['c']),
            volume=pair.round_quantity(data['v']),
            pair=pair
        )


@dataclass
class OrderInBook:
    price: float
    quantity: float
    count: int
    pair: Pair
    side: OrderSide

    @property
    def volume(self) -> float:
        return self.pair.round_quantity(self.price * self.quantity)


@dataclass
class OrderBook:
    buys: List[OrderInBook]
    sells: List[OrderInBook]
    pair: Pair

    @property
    def spread(self) -> float:
        return round_down(self.sells[0].price / self.buys[0].price - 1, 6)


@dataclass
class Balance:
    total: float
    available: float
    in_orders: float
    in_stake: float
    coin: Coin

    @classmethod
    def from_api(cls, data):
        return cls(
            total=data['balance'],
            available=data['available'],
            in_orders=data['order'],
            in_stake=data['stake'],
            coin=Coin(data['currency'])
        )


class OrderType(str, Enum):
    LIMIT = 'LIMIT'
    MARKET = 'MARKET'
    STOP_LOSS = 'STOP_LOSS'
    STOP_LIMIT = 'STOP_LIMIT'
    TAKE_PROFIT = 'TAKE_PROFIT'
    TAKE_PROFIT_LIMIT = 'TAKE_PROFIT_LIMIT'


class OrderStatus(str, Enum):
    ACTIVE = 'ACTIVE'
    FILLED = 'FILLED'
    CANCELED = 'CANCELED'
    REJECTED = 'REJECTED'
    EXPIRED = 'EXPIRED'
    PENDING = 'PENDING'


class OrderExecType(str, Enum):
    MARKET = ''
    POST_ONLY = 'POST_ONLY'


class OrderForceType(str, Enum):
    GOOD_TILL_CANCEL = 'GOOD_TILL_CANCEL'
    FILL_OR_KILL = 'FILL_OR_KILL'
    IMMEDIATE_OR_CANCEL = 'IMMEDIATE_OR_CANCEL'


@dataclass
class PrivateTrade:
    id: int
    side: OrderSide
    pair: Pair
    fees: float
    fees_coin: Coin
    created_at: int
    filled_price: float
    filled_quantity: float
    order_id: int

    @cached_property
    def is_buy(self):
        return self.side == OrderSide.BUY

    @cached_property
    def is_sell(self):
        return self.side == OrderSide.SELL

    @classmethod
    def create_from_api(cls, pair: Pair, data: Dict) -> 'PrivateTrade':
        return cls(
            id=int(data['trade_id']),
            side=OrderSide(data['side']),
            pair=pair,
            fees=round_up(data['fee'], 8),
            fees_coin=Coin(data['fee_currency']),
            created_at=int(data['create_time'] / 1000),
            filled_price=pair.round_price(data['traded_price']),
            filled_quantity=pair.round_quantity(data['traded_quantity']),
            order_id=int(data['order_id'])
        )


@dataclass
class Order:
    id: int
    status: OrderStatus
    side: OrderSide
    price: float
    quantity: float
    client_id: str
    created_at: int
    updated_at: int
    type: OrderType
    pair: Pair
    filled_quantity: float
    filled_price: float
    fees_coin: Coin
    force_type: OrderForceType
    trigger_price: float
    trades: List[PrivateTrade]

    @cached_property
    def is_buy(self):
        return self.side == OrderSide.BUY

    @cached_property
    def is_sell(self):
        return self.side == OrderSide.SELL

    @cached_property
    def is_active(self):
        return self.status == OrderStatus.ACTIVE

    @cached_property
    def is_canceled(self):
        return self.status == OrderStatus.CANCELED

    @cached_property
    def is_rejected(self):
        return self.status == OrderStatus.REJECTED

    @cached_property
    def is_expired(self):
        return self.status == OrderStatus.EXPIRED

    @cached_property
    def is_filled(self):
        return self.status == OrderStatus.FILLED

    @cached_property
    def is_pending(self):
        return self.status == OrderStatus.PENDING

    @cached_property
    def volume(self):
        return self.pair.round_quantity(self.price * self.quantity)

    @cached_property
    def filled_volume(self):
        return self.pair.round_quantity(
            self.filled_price * self.filled_quantity)

    @cached_property
    def remain_volume(self):
        return self.pair.round_quantity(self.filled_volume - self.volume)

    @cached_property
    def remain_quantity(self):
        return self.pair.round_quantity(self.quantity - self.filled_quantity)

    @classmethod
    def create_from_api(
            cls, pair: Pair, data: Dict, trades: List[Dict] = None) -> 'Order':
        fees_coin, trigger_price = None, None
        if data['fee_currency']:
            fees_coin = Coin(data['fee_currency'])
        if data.get('trigger_price') is not None:
            trigger_price = pair.round_price(data['trigger_price'])

        trades = [
            PrivateTrade.create_from_api(pair, trade)
            for trade in trades or []
        ]

        return cls(
            id=int(data['order_id']),
            status=OrderStatus(data['status']),
            side=OrderSide(data['side']),
            price=pair.round_price(data['avg_price'] or data['price']),
            quantity=pair.round_quantity(data['quantity']),
            client_id=data['client_oid'],
            created_at=int(data['create_time'] / 1000),
            updated_at=int(data['update_time'] / 1000),
            type=OrderType(data['type']),
            pair=pair,
            filled_price=pair.round_price(data['avg_price']),
            filled_quantity=pair.round_quantity(data['cumulative_quantity']),
            fees_coin=fees_coin,
            force_type=OrderForceType(data['time_in_force']),
            trigger_price=trigger_price,
            trades=trades
        )
